""" Testing File Validation Methods
"""
from pathlib import Path
from unittest.mock import Mock

import pytest

from changelist_data.storage import StorageType
from changelist_data.storage.file_validation import validate_file_input_text, file_exists, check_if_default_file_exists


def test_file_exists_does_not_exist_returns_false():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: False)
        assert not file_exists(Path('any'))


def test_file_exists_is_not_file_returns_false():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: False)
        assert not file_exists(Path('any'))


def test_file_exists_is_file_returns_true():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        assert file_exists(Path('any'))


def test_check_if_default_file_exists_changelists_does_not_exist_returns_none():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: False)
        assert check_if_default_file_exists(StorageType.CHANGELISTS) is None


def test_check_if_default_file_exists_changelists_is_not_file_returns_none():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: False)
        assert check_if_default_file_exists(StorageType.CHANGELISTS) is None


def test_check_if_default_file_exists_changelists_is_file_returns_path():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        assert isinstance(
            check_if_default_file_exists(StorageType.CHANGELISTS), Path
        )


def test_check_if_default_file_exists_workspace_does_not_exist_returns_none():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: False)
        assert check_if_default_file_exists(StorageType.WORKSPACE) is None


def test_check_if_default_file_exists_workspace_is_not_file_returns_none():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: False)
        assert check_if_default_file_exists(StorageType.WORKSPACE) is None


def test_check_if_default_file_exists_workspace_is_file_returns_true():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        assert isinstance(
            check_if_default_file_exists(StorageType.WORKSPACE), Path
        )


def test_validate_file_input_text_does_not_exist_raises_exit():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: False)
        try:
            validate_file_input_text(Path("file_name"))
            raises_exit = False
        except SystemExit:
            raises_exit = True
        assert raises_exit


def test_validate_file_input_text_is_not_file_raises_exit():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: False)
        try:
            validate_file_input_text(Path("file_name"))
            raises_exit = False
        except SystemExit:
            raises_exit = True
        assert raises_exit


def test_validate_file_input_text_filesize_too_large_raises_exit():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        obj = Mock()
        obj.__dict__["st_size"] = 33 * 1024 ** 3
        c.setattr(Path, 'stat', lambda _: obj)
        try:
            validate_file_input_text(Path("file_name"))
            raises_exit = False
        except SystemExit:
            raises_exit = True
        assert raises_exit


def test_validate_file_input_text_simple_changelist_returns_str(simple_changelists_xml):
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        obj = Mock()
        obj.__dict__["st_size"] = 4 * 1024
        c.setattr(Path, 'stat', lambda _: obj)
        c.setattr(Path, 'read_text', lambda _: simple_changelists_xml)
        simple_xml = validate_file_input_text(Path("file_name"))
    assert len(simple_xml) == 259



""" Changelist Data Storage XML Changelists Data File.
"""
from changelist_data.changelist import Changelist
from changelist_data.xml.changelists import changelists_reader
from changelist_data.xml.changelists.changelists_tree import ChangelistsTree


EMPTY_CHANGELISTS_DATA = """<?xml version="1.0" encoding="UTF-8"?>
<changelists>
</changelists>"""


def read_xml(
    changelists_xml: str
) -> list[Changelist]:
    """
    Parse the ChangeLists XML file and obtain all ChangeList Data in a list.

    Parameters:
    - changelists_xml (str): The contents of the ChangeLists file, in xml format.
    
    Returns:
    list[Changelist] - The list of Changelist objects in the ChangeLists file.
    """
    if (cl_manager := changelists_reader.find_changelists_root(changelists_reader.parse_xml(changelists_xml))) is None:
        exit("Changelists tag was not found in the xml file.")
    return changelists_reader.extract_list_elements(cl_manager)


def load_xml(
    changelists_xml: str
) -> ChangelistsTree:
    """
    Parse the Changelists XML file into an XML Tree, and Wrap it.

    Returns:
    ChangelistsTree - An XML Tree changelists interface.
    """
    return ChangelistsTree(
        changelists_reader.parse_xml(changelists_xml)
    )


def new_tree() -> ChangelistsTree:
    """
    Create a new Changelists XML Tree, and Wrap it.

    Returns:
    ChangelistsTree - An XML Tree changelists interface.
    """
    return ChangelistsTree(
        changelists_reader.parse_xml(EMPTY_CHANGELISTS_DATA)
    )



""" Reads the Workspace File, translates into Changelist Data types.
"""
from xml.etree.ElementTree import Element, ParseError, fromstring

from changelist_data.file_change import FileChange
from changelist_data.changelist import Changelist
from changelist_data.xml import xml_reader


def parse_xml(workspace_xml: str) -> Element:
    """ Parse an XML File. This should be a Workspace XML file.
        Returns the XML Root Element, or raises SystemExit.
    """
    try:
        return fromstring(workspace_xml)
    except ParseError:
        exit("Unable to Parse Workspace XML File.")


def find_changelist_manager(
    xml_root: Element
) -> Element | None:
    """ Find the ChangeList Manager in the XML Element Hierarchy.
        Looks for a Component Tag with the right name attribute.
        Returns None if the Element is not found.
    """
    for elem in xml_reader.filter_by_tag(xml_root, 'component'):
        try:
            if elem.attrib["name"] == 'ChangeListManager':
                return elem
        except KeyError:
            pass
    return None


def extract_list_elements(
    changelist_manager: Element
) -> list[Changelist]:
    """
    Given the Changelist Manager Element, obtain the list of List Elements.

    Parameters:
    - changelist_manager (Element): The ChangeList Manager XML Element.

    Returns:
    list[Element] - A List containing the Lists.
    """
    return [
        Changelist(
            id=xml_reader.get_attr(cl_element, 'id'),
            name=xml_reader.get_attr(cl_element, 'name'),
            changes=_extract_change_data(cl_element),
            comment=xml_reader.get_attr_or(cl_element, 'comment', ''),
            is_default=xml_reader.read_bool_from(cl_element, 'default'),
        ) for cl_element in xml_reader.filter_by_tag(changelist_manager, 'list')
    ]


_PROJECT_DIR_VAR = '$PROJECT_DIR$'
_PROJECT_DIR_LEN = len(_PROJECT_DIR_VAR)


def _filter_project_dir(path_str: str | None) -> str:
    """Filter the ProjectDir string at the beginning of the path.
    """
    if path_str is None:
        return None
    if path_str.startswith(_PROJECT_DIR_VAR):
        return path_str[_PROJECT_DIR_LEN:]
    return path_str


def _extract_change_data(
    list_element: Element,
) -> list[FileChange]:
    """
    Given a ChangeList XML Element, obtain the List of Changes.

    Parameters:
    - list_element (Element):  The Element representing a Changelist.

    Returns:
    list[FileChange] - The list of structured FileChange.
    """
    return [
        FileChange(
            before_path=_filter_project_dir(xml_reader.get_attr(change, 'beforePath')),
            before_dir=xml_reader.convert_bool(xml_reader.get_attr(change, 'beforeDir')),
            after_path=_filter_project_dir(xml_reader.get_attr(change, 'afterPath')),
            after_dir=xml_reader.convert_bool(xml_reader.get_attr(change, 'afterDir')),
        ) for change in filter(lambda x: x.tag == 'change', list_element)
    ]



"""
"""
import pytest

from changelist_data.changelist import Changelist
from changelist_data.file_change import FileChange


MODULE_SRC_PATH = '/module/src/main/java/module/Main.java'


@pytest.fixture()
def fc_all():
    return FileChange(
        before_path=MODULE_SRC_PATH,
        after_path=MODULE_SRC_PATH,
        before_dir=False,
        after_dir=False
    )


@pytest.fixture()
def fc_before():
    return FileChange(
        before_path=MODULE_SRC_PATH,
        before_dir=False,
    )


@pytest.fixture()
def fc_after():
    return FileChange(
        after_path=MODULE_SRC_PATH,
        after_dir=False
    )


@pytest.fixture()
def simple_cl():
    return Changelist(
        id="9f60fda2-421e-4a4b-bd0f-4c8f83a47c88",
        name="Simple",
        changes=[
            FileChange(before_path='/main.py',
                       before_dir=False,
                       after_path='/main.py',
                       after_dir=False)
        ],
        comment="Main Program Files",
    )


@pytest.fixture()
def simple_cl_list(simple_cl):
    return [simple_cl]


@pytest.fixture()
def multi_cl_list():
    return [
        Changelist(
            id="af84ea1b-1b24-407d-970f-9f3a2835e933",
            name="Main",
            changes=[
                FileChange(before_path='/history.py',
                           before_dir=False,
                           after_path=None,
                           after_dir=None),
                FileChange(before_path='/main.py',
                           before_dir=False,
                           after_path=None,
                           after_dir=None)
            ],
            comment="Main Program Files",
            is_default=True,
        ),
        Changelist(
            id="9f60fda2-421e-4a4b-bd0f-4c8f83a47c88",
            name="Test",
            changes=[
                FileChange(before_path=None,
                           before_dir=None,
                           after_path='/test/test_file.py',
                           after_dir=False)
            ],
            comment="Test Files",
        ),
    ]


@pytest.fixture()
def no_clm_workspace_xml() -> str:
    """Workspace XML with no ChangeListManager component.
    """
    return """<?xml version="1.0" encoding="UTF-8"?>
<project version="4">
  <component name="AutoImportSettings">
    <option name="autoReloadType" value="NONE" />
  </component>
</project>"""


@pytest.fixture()
def empty_workspace_xml() -> str:
    """Workspace XML with empty ChangeListManager component.
    """
    return """<?xml version="1.0" encoding="UTF-8"?>
<project version="4">
  <component name="AutoImportSettings">
    <option name="autoReloadType" value="NONE" />
  </component>
  <component name="ChangeListManager">
  </component>
</project>"""


@pytest.fixture()
def simple_workspace_xml() -> str:
    """Workspace XML with a Simple Changelist.
    """
    return """<?xml version="1.0" encoding="UTF-8"?>
<project version="4">
  <component name="AutoImportSettings">
    <option name="autoReloadType" value="NONE" />
  </component>
  <component name="ChangeListManager">
    <list id="9f60fda2-421e-4a4b-bd0f-4c8f83a47c88" name="Simple" comment="Main Program Files">
      <change beforePath="$PROJECT_DIR$/main.py" beforeDir="false" afterPath="$PROJECT_DIR$/main.py" afterDir="false" />
    </list>
  </component>
</project>"""


@pytest.fixture()
def multi_workspace_xml():
    """Workspace XML with Multiple Changelists.
    """
    return """<?xml version="1.0" encoding="UTF-8"?>
<project version="4">
  <component name="AutoImportSettings">
    <option name="autoReloadType" value="NONE" />
  </component>
  <component name="ChangeListManager">
    <list default="true" id="af84ea1b-1b24-407d-970f-9f3a2835e933" name="Main" comment="Main Program Files">
      <change beforePath="$PROJECT_DIR$/history.py" beforeDir="false" />
      <change beforePath="$PROJECT_DIR$/main.py" beforeDir="false" />
    </list>
    <list id="9f60fda2-421e-4a4b-bd0f-4c8f83a47c88" name="Test" comment="Test Files">
      <change afterPath="$PROJECT_DIR$/test/test_file.py" afterDir="false" />
    </list>
  </component>
</project>"""


@pytest.fixture()
def empty_changelists_xml() -> str:
    return """<?xml version="1.0" encoding="UTF-8"?>
<changelists></changelists>"""


@pytest.fixture()
def simple_changelists_xml() -> str:
    return """<?xml version="1.0" encoding="UTF-8"?>
<changelists>
<list id="9f60fda2-421e-4a4b-bd0f-4c8f83a47c88" name="Simple" comment="Main Program Files">
  <change beforePath="/main.py" beforeDir="false"  afterPath="/main.py" afterDir="false" />
</list>
</changelists>"""


@pytest.fixture()
def multi_changelists_xml() -> str:
    return """<?xml version="1.0" encoding="UTF-8"?>
<changelists>
<list default="true" id="af84ea1b-1b24-407d-970f-9f3a2835e933" name="Main" comment="Main Program Files">
  <change beforePath="/history.py" beforeDir="false" />
  <change beforePath="/main.py" beforeDir="false" />
</list>
<list id="9f60fda2-421e-4a4b-bd0f-4c8f83a47c88" name="Test" comment="Test Files">
  <change afterPath="/test/test_file.py" afterDir="false" />
</list>
</changelists>"""


@pytest.fixture()
def invalid_changelists_xml() -> str:
    return """<?xml version="1.0" encoding="UTF-8"?>
<changelists>
<lisat id="9f60fda2-421e-4a4b-bd0f-4c8f83a47c88" name="Main" comment="Main Files">
  <change beforePath="/main.py" beforeDir="false" />
</lisat>
</changelists>"""



""" Testing Workspace Writer
"""
from changelist_data.file_change import FileChange
from changelist_data.xml.workspace import workspace_reader
from changelist_data.xml.workspace.workspace_writer import _write_change_data, write_list_element


def test_write_change_data_with_no_fields():
    element = _write_change_data(FileChange(), 1)
    assert element.tag == 'change'
    assert element.attrib == {}


def test_write_change_data_with_all_fields(fc_all):
    element = _write_change_data(fc_all, 1)
    assert element.tag == 'change'
    assert element.get('beforePath') == f'{workspace_reader._PROJECT_DIR_VAR}{fc_all.before_path}'
    assert element.get('afterPath') == f'{workspace_reader._PROJECT_DIR_VAR}{fc_all.after_path}'
    assert element.get('beforeDir') == 'false'
    assert element.get('afterDir') == 'false'


def test_write_change_data_with_before_fields(fc_before):
    element = _write_change_data(fc_before, 1)
    assert element.tag == 'change'
    assert element.get('beforePath') == f'{workspace_reader._PROJECT_DIR_VAR}{fc_before.before_path}'
    assert element.get('beforeDir') == 'false'
    assert element.get('afterDir') is None
    assert element.get('afterPath') is None


def test_write_change_data_with_after_fields(fc_after):
    element = _write_change_data(fc_after, 1)
    assert element.tag == 'change'
    assert element.get('beforePath') is None
    assert element.get('afterPath') == f'{workspace_reader._PROJECT_DIR_VAR}{fc_after.after_path}'
    assert element.get('afterDir') == 'false'
    assert element.get('beforeDir') is None


def test_write_list_element_simple_returns(simple_cl):
    result = write_list_element(simple_cl, 2)
    assert result.tag == 'list'


def test_write_list_element_multi_returns(multi_cl_list):
    result0 = write_list_element(multi_cl_list[0], 2)
    result1 = write_list_element(multi_cl_list[1], 2)
    assert result0.tag == 'list'
    assert result1.tag == 'list'
    assert len(list(result0.iter('change'))) == 2
    assert len(list(result1.iter('change'))) == 1



""" Testing Changelists Reader.
"""
from changelist_data.xml.changelists import read_xml


def test_read_xml_empty_string_raises_exit():
    try:
        read_xml("")
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit


def test_read_xml_no_cl_returns_empty_list(empty_changelists_xml):
    result = read_xml(empty_changelists_xml)
    assert len(result) == 0

    
def test_read_xml_simple_xml_returns_simple_cl(simple_changelists_xml, simple_cl):
    assert [simple_cl] == read_xml(simple_changelists_xml)


def test_read_xml_multi_xml_returns_multi_cl_list(multi_changelists_xml, multi_cl_list):
    assert multi_cl_list == read_xml(multi_changelists_xml)



"""Data describing a File Change.
"""
from dataclasses import dataclass


@dataclass(frozen=True)
class FileChange:
    """The Change Information that is associated with a single file.

    Properties:
    - before_path (str | None): The initial path of the file.
    - before_dir (bool | None): Whether the initial file is a directory.
    - after_path (str | None): The final path of the file.
    - after_dir (bool | None): Whether the final path is a directory.
    """
    before_path: str | None = None
    before_dir: bool | None = None
    after_path: str | None = None
    after_dir: bool | None = None



""" Testing Changelists XML load_xml Method.
"""
from changelist_data.xml.changelists import load_xml


def test_load_xml_empty_str_raises_exit():
    try:
        load_xml("")
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit


def test_load_xml_empty_returns_empty(empty_changelists_xml):
    result = load_xml(empty_changelists_xml).get_changelists()
    assert len(result) == 0


def test_load_xml_simple_cl(simple_changelists_xml, simple_cl_tree):
    result = load_xml(simple_changelists_xml).get_changelists()
    assert len(result) == 1
    assert simple_cl_tree.get_changelists() == result


def test_load_xml_multi_cl(multi_changelists_xml, multi_cl_tree):
    result = load_xml(multi_changelists_xml).get_changelists()
    assert len(result) == 2
    assert multi_cl_tree.get_changelists() == result


"""Testing ChangeList Data
"""
import pytest

from changelist_data.changelist import Changelist
from changelist_data.file_change import FileChange


@pytest.fixture()
def sample_cl0():
    return Changelist(
        id="0",
        name="",
        changes=list(),
    )


@pytest.fixture()
def sample_cl1():
    return Changelist(
        id="1212434",
        name="ChangeList",
        changes=[
            FileChange(
                after_path="/module/file.txt",
                after_dir=False,
            )
        ],
    )


def test_properties_cl0(sample_cl0):
    assert sample_cl0.id == '0'
    assert sample_cl0.name == ''
    assert len(sample_cl0.changes) == 0
    assert len(sample_cl0.comment) == 0
    assert not sample_cl0.is_default


def test_properties_cl1(sample_cl1):
    assert sample_cl1.id == '1212434'
    assert sample_cl1.name == 'ChangeList'
    assert len(sample_cl1.changes) == 1
    assert len(sample_cl1.comment) == 0
    assert not sample_cl1.is_default



""" Testing Changelist Data Storage Package Methods.
"""
from pathlib import Path
from unittest.mock import Mock

import pytest

from changelist_data.storage import read_storage
from changelist_data.storage.storage_type import StorageType


def test_read_storage_defaults_no_files_exist_returns_empty_list():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: False)
        result = read_storage()
    assert len(result) == 0


def test_read_storage_defaults_all_paths_are_dirs_returns_empty_list():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: False)
        result = read_storage()
    assert len(result) == 0


def test_read_storage_default_simple_cl_(simple_changelists_xml):
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        obj = Mock()
        obj.__dict__["st_size"] = 4 * 1024
        c.setattr(Path, 'stat', lambda _: obj)
        c.setattr(Path, 'read_text', lambda _: simple_changelists_xml)
        result = read_storage()
    assert len(result) == 1


def test_read_storage_file_does_not_exist_():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: False)
        try:
            read_storage(None, Path("filepath"))
            raises_exit = False
        except SystemExit:
            raises_exit = True
        assert raises_exit


def test_read_storage_changelists_simple_cl_(simple_changelists_xml):
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        obj = Mock()
        obj.__dict__["st_size"] = 4 * 1024
        c.setattr(Path, 'stat', lambda _: obj)
        c.setattr(Path, 'read_text', lambda _: simple_changelists_xml)
        result = read_storage(StorageType.CHANGELISTS, Path("filepath"))
    assert len(result) == 1


def test_read_storage_workspace_simple_cl_(simple_workspace_xml):
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        obj = Mock()
        obj.__dict__["st_size"] = 4 * 1024
        c.setattr(Path, 'stat', lambda _: obj)
        c.setattr(Path, 'read_text', lambda _: simple_workspace_xml)
        result = read_storage(StorageType.WORKSPACE, Path("filepath"))
    assert len(result) == 1


def test_read_storage_changelists_simple_cl_returns_simple_list(simple_changelists_xml):
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        obj = Mock()
        obj.__dict__["st_size"] = 4 * 1024
        c.setattr(Path, 'stat', lambda _: obj)
        c.setattr(Path, 'read_text', lambda _: simple_changelists_xml)
        result = read_storage(StorageType.CHANGELISTS)
    assert len(result) == 1


def test_read_storage_changelists_multi_cl_returns_list(multi_changelists_xml):
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        obj = Mock()
        obj.__dict__["st_size"] = 4 * 1024
        c.setattr(Path, 'stat', lambda _: obj)
        c.setattr(Path, 'read_text', lambda _: multi_changelists_xml)
        result = read_storage(StorageType.CHANGELISTS)
    assert len(result) == 2



def test_read_storage_workspace_simple_cl_returns_simple_list(simple_workspace_xml):
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        obj = Mock()
        obj.__dict__["st_size"] = 4 * 1024
        c.setattr(Path, 'stat', lambda _: obj)
        c.setattr(Path, 'read_text', lambda _: simple_workspace_xml)
        result = read_storage(StorageType.WORKSPACE)
    assert len(result) == 1


def test_read_storage_workspace_multi_cl_returns_list(multi_workspace_xml):
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        obj = Mock()
        obj.__dict__["st_size"] = 4 * 1024
        c.setattr(Path, 'stat', lambda _: obj)
        c.setattr(Path, 'read_text', lambda _: multi_workspace_xml)
        result = read_storage(StorageType.WORKSPACE)
    assert len(result) == 2



""" Storage for Changelists Data XML Format.
"""
from pathlib import Path

from changelist_data.changelist import Changelist
from changelist_data.storage import storage_type, file_validation
from changelist_data.storage.storage_type import StorageType
from changelist_data.xml import changelists
from changelist_data.xml.changelists import ChangelistsTree, EMPTY_CHANGELISTS_DATA


def read_file(
    file_path: Path | None = storage_type.get_default_path(StorageType.CHANGELISTS),
) -> list[Changelist]:
    """ Read a Changelists XML Storage File.
        Default file_path is given by StorageType.
        None file_path returns an empty list.

    Parameters:
    - file_path (Path): The Path to the File containing Changelists XML.

    Returns:
    list[Changelist] - The list of Changelist data stored in Changelists Storage.
    """
    if file_path is None:
        return [] # None Path Given
    if len(file_content := file_validation.validate_file_input_text(file_path)) == 0:
        return [] # Empty Data File
    return changelists.read_xml(file_content)


def load_file(
    file_path: Path | None = storage_type.get_default_path(StorageType.CHANGELISTS),
) -> ChangelistsTree:
    """ Load a Tree from Changelists XML Storage File.
        Default file_path

    Parameters:
    - file_path (Path): The Path to the File containing Changelists XML.

    Returns:
    ChangelistsTree - The list of Changelist data stored in Changelists Storage.
    """
    if file_path is None or\
        not file_validation.file_exists(file_path):
        return changelists.new_tree()
    # Validate File Stats, Read and Parse the XML
    if len(file_content := file_validation.validate_file_input_text(file_path)) == 0:
        return changelists.new_tree() # Empty Data File
    # Parse and wrap XML ElementTree in ChangelistsTree
    return changelists.load_xml(file_content)


def write_file(
    tree: ChangelistsTree | None,
    file_path: Path = storage_type.get_default_path(StorageType.CHANGELISTS),
) -> bool:
    """ Write a Changelist Data Storage object to an XML File.

    Parameters:
    - tree (ChangelistDataStorage): The Tree object containing the Changelists.
    - file_path (Path): The File Path to write the XML data to.

    Returns:
    bool - True after the operation succeeds.
    """
    if tree is None:
        file_path.write_text(EMPTY_CHANGELISTS_DATA)
        return True
    tree.write_tree(file_path)
    return True



""" Testing Storage of Changelists XML Files.
"""
from pathlib import Path

from changelist_data.storage import StorageType
from changelist_data.storage.storage_type import CHANGELISTS_FILE_PATH_STR, get_default_path, WORKSPACE_FILE_PATH_STR


def test_get_default_path_changelist():
    assert Path(CHANGELISTS_FILE_PATH_STR) == get_default_path(StorageType.CHANGELISTS)


def test_get_default_path_workspace():
    assert Path(WORKSPACE_FILE_PATH_STR) == get_default_path(StorageType.WORKSPACE)



""" Testing the WorkspaceTree Class.
"""


def test_get_changelists_simple_returns_list(simple_ws_tree):
    result = simple_ws_tree.get_changelists()
    assert len(result) == 1
    cl = result[0]
    assert cl.name == 'Simple'
    assert cl.comment == 'Main Program Files'
    assert len(cl.changes) == 1
    file = cl.changes[0]
    assert file.before_path == '/main.py'
    assert not file.before_dir
    assert file.after_path == '/main.py'
    assert not file.after_dir


def test_get_changelists_multi_returns_list(multi_ws_tree):
    result = multi_ws_tree.get_changelists()
    assert len(result) == 2
    # First Changelist
    cl_0 = result[0]
    assert cl_0.name == 'Main'
    assert len(cl_0.changes) == 2
    # Second Changelist
    cl_1 = result[1]
    assert cl_1.name == 'Test'
    assert len(cl_1.changes) == 1


def test_get_changelists_no_cl_returns_empty_list(no_clm_ws_tree):
    try:
        no_clm_ws_tree.get_changelists()
        assert False
    except SystemExit:
        assert True


def test_update_changelists_simple_with_empty_returns_empty(simple_ws_tree):
    simple_ws_tree.update_changelists([])
    assert len(simple_ws_tree.get_changelists()) == 0


def test_update_changelists_multi_with_empty_returns_empty(multi_ws_tree):
    multi_ws_tree.update_changelists([])
    assert len(multi_ws_tree.get_changelists()) == 0


def test_update_changelists_simple_with_multi(simple_ws_tree, multi_ws_tree):
    simple_ws_tree.update_changelists(
        multi_ws_tree.get_changelists()
    )
    assert simple_ws_tree.get_changelists() == multi_ws_tree.get_changelists()


def test_update_changelists_multi_with_simple(multi_ws_tree, simple_ws_tree):
    multi_ws_tree.update_changelists(
        simple_ws_tree.get_changelists()
    )
    assert multi_ws_tree.get_changelists() == simple_ws_tree.get_changelists()


def test_update_changelists_no_cl_manager_raises_exit(no_clm_ws_tree):
    try:
        no_clm_ws_tree.update_changelists([])
        raised_exit = False
    except SystemExit:
        raised_exit = True
    assert raised_exit


def test_get_root_no_cl_manager_returns_root(no_clm_ws_tree):
    assert no_clm_ws_tree.get_root().getroot().tag == 'project'


def test_get_root_simple_returns_root(simple_ws_tree):
    assert simple_ws_tree.get_root().getroot().tag == 'project'


def test_get_root_multi_returns_root(multi_ws_tree):
    assert multi_ws_tree.get_root().getroot().tag == 'project'



"""Testing Module Methods.
"""
from test.changelist_data.conftest import MODULE_SRC_PATH


def test_properties_module_src_after(fc_after):
    data = fc_after
    assert not data.before_dir 
    assert data.before_path is None
    assert not data.after_dir
    assert data.after_path == MODULE_SRC_PATH


def test_properties_module_src_before(fc_before):
    data = fc_before
    assert not data.before_dir
    assert data.before_path == MODULE_SRC_PATH
    assert not data.after_dir
    assert data.after_path is None


def test_properties_module_src_all(fc_all):
    data = fc_all
    assert not data.before_dir
    assert data.before_path == MODULE_SRC_PATH
    assert not data.after_dir
    assert data.after_path == MODULE_SRC_PATH



"""
"""
from xml.etree.ElementTree import fromstring

import pytest

from changelist_data.xml.changelists import ChangelistsTree


@pytest.fixture()
def simple_cl_tree(simple_changelists_xml):
    return ChangelistsTree(fromstring(simple_changelists_xml))


@pytest.fixture()
def multi_cl_tree(multi_changelists_xml):
    return ChangelistsTree(fromstring(multi_changelists_xml))


@pytest.fixture()
def no_cl_tree(empty_changelists_xml):
    return ChangelistsTree(fromstring(empty_changelists_xml))


@pytest.fixture()
def invalid_cl_xml() -> str:
    return """<?xml version="1.0" encoding="UTF-8"?>
<changelists>
<lisat id="af84ea1b9f3a2835e933" name="Main" comment="Main Files">
  <change beforePath="/main.py" beforeDir="false" />
</lisat>
</changelists>"""



""" XML Tree Class for Changelists Data XML.
"""
from xml.etree.ElementTree import Element, ElementTree, indent

from changelist_data.changelist import Changelist
from changelist_data.xml.base_xml_tree import BaseXMLTree
from changelist_data.xml.changelists import changelists_reader, changelists_writer


class ChangelistsTree(BaseXMLTree):
    """
    Manages the Changelists Data XML Element Trees.

    Properties:
    - xml_root (Element): The XML root element.
    - changelists_element (Element): The Changelists Tag Element.
    """
    
    def __init__(
        self,
        xml_root: Element,
    ):
        self._xml_root = xml_root
        self.changelists_element = changelists_reader.find_changelists_root(xml_root)

    def get_changelists(self) -> list[Changelist]:
        """
        Given the Changelist Manager Element, obtain the list of List Elements.

        Returns:
        list[Changelist] - A List containing the Lists.
        """
        if self.changelists_element is None:
            exit('XML File does not have a Changelists Element.')
        return changelists_reader.extract_list_elements(self.changelists_element)

    def get_root(self) -> ElementTree:
        """
        Obtain the XML ElementTree Root.
        """
        return ElementTree(self._xml_root)

    def update_changelists(
        self,
        changelists: list[Changelist],
    ):
        """
        Update the XML Tree's Changelists.
        
        Parameters:
        - changelists (list[Changelist]): The List of Changelists.
        """
        clm = self.changelists_element
        if clm is None:
            exit('XML File does not have a Changelists Element.')
        # Clear the Changelists Element
        clm.clear() # Need to Restore any Attributes after Clear operation:
        # - No attributes in use by changelists at the moment.
        # Add All Sub Elements
        clm.extend(changelists_writer.write_list_element(x) for x in changelists)
        indent(clm, level=0)



""" Testing Workspace XML Package Methods.
"""
from changelist_data.xml.workspace import read_xml


def test_read_xml_empty_str_raises_exit():
    try:
        read_xml("")
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit


def test_read_xml_no_cl_raises_exit(no_clm_workspace_xml):
    try:
        read_xml(no_clm_workspace_xml)
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit


def test_read_xml_simple_cl(simple_workspace_xml, simple_cl_list):
    assert simple_cl_list == read_xml(simple_workspace_xml)


def test_read_xml_multi_cl(multi_workspace_xml, multi_cl_list):
    assert multi_cl_list == read_xml(multi_workspace_xml)



""" Testing Storage of Workspace XML Files.
"""
from changelist_data.storage import workspace_storage


def test_read_file_no_cl_raises_exit(temp_file, no_clm_workspace_xml):
    temp_file.write_text(no_clm_workspace_xml)
    try:
        result = workspace_storage.read_file(temp_file)
        assert len(result) == 0
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit


def test_read_file_empty_cl(temp_file, empty_workspace_xml):
    temp_file.write_text(empty_workspace_xml)
    result = workspace_storage.read_file(temp_file)
    assert len(result) == 0


def test_read_file_simple_cl(temp_file, simple_workspace_xml):
    temp_file.write_text(simple_workspace_xml)
    result = workspace_storage.read_file(temp_file)
    assert len(result) == 1


def test_read_file_multi_cl(temp_file, multi_workspace_xml):
    temp_file.write_text(multi_workspace_xml)
    result = workspace_storage.read_file(temp_file)
    assert len(result) == 2


def test_load_file_no_cl(temp_file, no_clm_workspace_xml):
    temp_file.write_text(no_clm_workspace_xml)
    try:
        result = workspace_storage.load_file(temp_file)
        result.get_changelists()
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit


def test_load_file_empty_cl(temp_file, empty_workspace_xml):
    temp_file.write_text(empty_workspace_xml)
    result = workspace_storage.load_file(temp_file)
    assert len(result.get_changelists()) == 0


def test_load_file_simple_cl(temp_file, simple_workspace_xml, simple_cl_list):
    temp_file.write_text(simple_workspace_xml)
    assert simple_cl_list == workspace_storage.load_file(temp_file).get_changelists()


def test_load_file_multi_cl(temp_file, multi_workspace_xml, multi_cl_list):
    temp_file.write_text(multi_workspace_xml)
    assert multi_cl_list == workspace_storage.load_file(temp_file).get_changelists()


def test_write_file_empty_cl(temp_file, empty_workspace_xml):
    temp_file.write_text(empty_workspace_xml)
    tree = workspace_storage.load_file(temp_file)
    workspace_storage.write_file(tree, temp_file)
    # Read from Temp
    result = workspace_storage.read_file(temp_file)
    assert len(result) == 0


def test_write_file_simple_cl(temp_file, simple_workspace_xml, simple_cl_list):
    temp_file.write_text(simple_workspace_xml)
    tree = workspace_storage.load_file(temp_file)
    workspace_storage.write_file(tree, temp_file)
    # Read from Temp
    assert simple_cl_list == workspace_storage.read_file(temp_file)


def test_write_file_multi_cl(temp_file, multi_workspace_xml, multi_cl_list):
    temp_file.write_text(multi_workspace_xml)
    tree = workspace_storage.load_file(temp_file)
    workspace_storage.write_file(tree, temp_file)
    # Read from Temp
    assert multi_cl_list == workspace_storage.read_file(temp_file)



""" Testing Changelist Data Storage Class and Methods.
"""
import changelist_data.xml.changelists
from changelist_data.changelist import Changelist
from changelist_data.storage import StorageType, ChangelistDataStorage, load_storage


def test_write_to_storage_changelists_file_does_not_exist_returns_true(temp_file):
    temp_file.touch()
    result = load_storage(StorageType.CHANGELISTS, temp_file)
    assert isinstance(result, ChangelistDataStorage)
    assert result.storage_type == StorageType.CHANGELISTS
    assert len(result.get_changelists()) == 0
    #
    assert result.write_to_storage()
    assert len(temp_file.read_text()) > 0


def test_write_to_storage_after_update_changelists_file_does_not_exist_returns_true(temp_file):
    temp_file.touch()
    storage = load_storage(StorageType.CHANGELISTS, temp_file)
    #
    storage.update_changelists([Changelist('id', 'name', [])])
    assert storage.write_to_storage()
    #
    result = changelist_data.xml.changelists.read_xml(temp_file.read_text())
    assert len(result) == 1
    assert result[0].id == 'id'
    assert result[0].name == 'name'
    assert len(result[0].changes) == 0



""" Storage Access Methods.
"""
from pathlib import Path

from changelist_data.changelist import Changelist
from changelist_data.storage import file_validation, changelists_storage, workspace_storage
from changelist_data.storage.changelist_data_storage import ChangelistDataStorage
from changelist_data.storage.storage_type import StorageType, get_default_path
from changelist_data.xml.changelists import new_tree


def read_storage(
    option: StorageType | None = None,
    file_path: Path | None = None,
) -> list[Changelist]:
    """ Read Changelist Data from Storage into a List of Changelist data.
        None values indicate that the storage file should be searched for using default values.

    Parameters:
    - option (StorageType | None): The Storage Type describing the XML format. Use None to search all StorageTypes in order.
    - file_path (Path | None): The file path to read from. Use None to look in default storage locations.

    Returns:
    list[Changelist] - The List of Changelist data.
    """
    if file_path is not None:
        if option is None:
            exit("When File Path is Provided, The Storage Type must be Provided as well.")
        return _read_option(option, file_path)
    # Find and Read from Default File Paths
    if option is None:
        for opts in StorageType:
            if (storage_path := file_validation.check_if_default_file_exists(StorageType(opts))) is not None:
                return _read_option(StorageType(opts), storage_path)
    elif (storage_path := file_validation.check_if_default_file_exists(option)) is not None:
        return _read_option(option, storage_path)
    return []


def load_storage(
    option: StorageType | None = None,
    file_path: Path | None = None,
) -> ChangelistDataStorage:
    """ Load Changelist Data into a managed Storage object.
        None values indicate that the storage file should be searched for using default values.

    Parameters:
    - option (StorageType | None): The Storage Type describing the XML format. Use None to search all StorageTypes in order.
    - file_path (Path | None): The file path to read from. Use None to look in default storage locations.

    Returns:
    ChangelistDataStorage - A managed Storage object containing Changelist data.
    """
    if file_path is not None:
        if option is None:
            exit("When File Path is Provided, The Storage Type must be Provided as well.")
        return _load_option(option, file_path)
    # Find and Read from Default File Paths
    if option is None:
        for opts in StorageType:
            if (storage_path := file_validation.check_if_default_file_exists(StorageType(opts))) is not None:
                return _load_option(StorageType(opts), storage_path)
    elif (storage_path := file_validation.check_if_default_file_exists(option)) is not None:
        return _load_option(option, storage_path)
    # Create an empty Changelists Storage Tree with a default path
    return ChangelistDataStorage(
        new_tree(),
        StorageType.CHANGELISTS,
        get_default_path(StorageType.CHANGELISTS)
    )


def _read_option(
    option: StorageType,
    path: Path
) -> list[Changelist]:
    if option == StorageType.CHANGELISTS:
        return changelists_storage.read_file(path)
    if option == StorageType.WORKSPACE:
        return workspace_storage.read_file(path)
    raise ValueError("Invalid Storage Option")


def _load_option(
    option: StorageType,
    path: Path
) -> ChangelistDataStorage:
    if option == StorageType.CHANGELISTS:
        return ChangelistDataStorage(
            changelists_storage.load_file(path),
            option,
            path
        )
    if option == StorageType.WORKSPACE:
        return ChangelistDataStorage(
            workspace_storage.load_file(path),
            option,
            path
        )
    raise ValueError("Invalid Storage Option")



""" Testing Changelists Writer
"""
from changelist_data.file_change import FileChange
from changelist_data.xml.changelists.changelists_writer import _write_change_data

from test.changelist_data.conftest import MODULE_SRC_PATH


def test_write_change_data_with_no_fields():
    data = FileChange()
    element = _write_change_data(data, 1)
    assert element.tag == 'change'
    assert element.attrib == {}


def test_write_change_data_with_all_fields(fc_all):
    element = _write_change_data(fc_all, 1)
    assert element.tag == 'change'
    assert element.get('beforePath') == MODULE_SRC_PATH
    assert element.get('afterPath') == MODULE_SRC_PATH
    assert element.get('beforeDir') == 'false'
    assert element.get('afterDir') == 'false'


def test_write_change_data_with_before_fields(fc_before):
    element = _write_change_data(fc_before, 1)
    assert element.tag == 'change'
    assert element.get('beforePath') == MODULE_SRC_PATH
    assert element.get('beforeDir') == 'false'
    assert element.get('afterDir') is None
    assert element.get('afterPath') is None


def test_write_change_data_with_after_fields(fc_after):
    element = _write_change_data(fc_after, 1)
    assert element.tag == 'change'
    assert element.get('beforePath') is None
    assert element.get('afterPath') == MODULE_SRC_PATH
    assert element.get('afterDir') == 'false'
    assert element.get('beforeDir') is None



""" File Validation Methods
"""
from pathlib import Path

from changelist_data.storage import storage_type
from changelist_data.storage.storage_type import StorageType


CHANGELISTS_FILE_PATH_STR = '.changelists/data.xml'
WORKSPACE_FILE_PATH_STR = '.idea/workspace.xml'


def file_exists(path: Path) -> bool:
    """ Determines if Path exists and is_file.
    """
    return path.exists() and path.is_file()


def check_if_default_file_exists(
    option: StorageType,
) -> Path | None:
    """ Return the Path to the default file for this storage option iff it exists.

    Parameters:
    - option (StorageType): The StorageType option to check.

    Returns:
    Path | None - The Path to the Default Storage File, or None if the path is not a file that exists.
    """
    if file_exists(file_path := storage_type.get_default_path(option)):
        return file_path
    return None


def validate_file_input_text(file_path: Path) -> str:
    """
    Ensure that the File Exists, and is within reasonable size parameters.
        Read the File and return its string contents.

    Parameters:
    - file_path (Path): The Path to the Input File.

    Returns:
    str - The Text Contents of the Input File.

    Raises:
    SystemExit - When any of the validation conditions fails, or the file operation fails.
    """
    if not file_path.exists():
        exit("File did not exist")
    if not file_path.is_file():
        exit("Given Path was not a file")
    file_stats = file_path.stat()
    if (file_size := file_stats.st_size / 1024) > 32 * 1024:
        exit("Input File was larger than 32 MB. Refusing to read it.")
    try:
        return file_path.read_text()
    except FileNotFoundError:
        exit("Couldn't find the file, after checking that it exists.")
    except IOError:
        exit("IOError occurred while reading Input File")
    except Exception as e:
        exit(f"Unexpected Exception while reading InputFile(name={file_path.name}, fileSize={file_size} kb) Exception=({e})")



""" Testing Changelist Data Storage load_storage Method.
"""
from pathlib import Path
from unittest.mock import Mock

import pytest

import changelist_data.xml.changelists
from changelist_data.storage import load_storage
from changelist_data.storage.changelist_data_storage import ChangelistDataStorage
from changelist_data.storage.storage_type import StorageType


def test_load_storage_changelists_empty_file_returns_new_tree(temp_file):
    temp_file.write_text("")
    result = load_storage(StorageType.CHANGELISTS, temp_file)
    #
    assert isinstance(result, ChangelistDataStorage)
    assert len(result.get_changelists()) == 0


def test_load_storage_changelists_file_does_not_exist_returns_empty(temp_file):
    temp_file.unlink(missing_ok=True)
    result = load_storage(StorageType.CHANGELISTS, temp_file)
    assert isinstance(result, ChangelistDataStorage)
    assert result.storage_type == StorageType.CHANGELISTS
    assert len(result.get_changelists()) == 0


def test_load_storage_workspace_empty_file_raises_exit(temp_file):
    temp_file.write_text("")
    try:
        load_storage(StorageType.WORKSPACE, temp_file)
        raised_exit = False
    except SystemExit:
        raised_exit = True
    assert raised_exit


def test_load_storage_changelists_no_changelists_returns_empty_storage(temp_file):
    temp_file.write_text(changelist_data.xml.changelists.EMPTY_CHANGELISTS_DATA)
    result = load_storage(StorageType.CHANGELISTS, temp_file)
    assert isinstance(result, ChangelistDataStorage)
    assert len(result.get_changelists()) == 0


def test_load_storage_workspace_no_clm_raises_exit(temp_file, no_clm_workspace_xml):
    temp_file.write_text(no_clm_workspace_xml)
    try:
        result = load_storage(StorageType.WORKSPACE, temp_file)
        assert isinstance(result, ChangelistDataStorage)
        assert len(result.get_changelists()) == 0
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit


def test_load_storage_workspace_empty_changelists_returns_empty_storage(temp_file, empty_workspace_xml):
    temp_file.write_text(empty_workspace_xml)
    result = load_storage(StorageType.WORKSPACE, temp_file)
    assert isinstance(result, ChangelistDataStorage)
    assert len(result.get_changelists()) == 0


def test_load_storage_changelists_simple_cl_returns_simple_list(temp_file, simple_changelists_xml):
    temp_file.write_text(simple_changelists_xml)
    result = load_storage(StorageType.CHANGELISTS, temp_file)
    assert len(result.get_changelists()) == 1


def test_load_storage_workspace_simple_cl_returns_simple_list(temp_file, simple_workspace_xml):
    temp_file.write_text(simple_workspace_xml)
    result = load_storage(StorageType.WORKSPACE, temp_file)
    assert len(result.get_changelists()) == 1


def test_load_storage_changelists_multi_cl_returns_multi_list(temp_file, multi_changelists_xml):
    temp_file.write_text(multi_changelists_xml)
    result = load_storage(StorageType.CHANGELISTS, temp_file)
    assert len(result.get_changelists()) == 2


def test_load_storage_workspace_multi_cl_returns_multi_list(temp_file, multi_workspace_xml):
    temp_file.write_text(multi_workspace_xml)
    result = load_storage(StorageType.WORKSPACE, temp_file)
    assert len(result.get_changelists()) == 2


def test_load_storage_none_changelists_simple_cl_returns_simple_list(simple_changelists_xml):
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        obj = Mock()
        obj.__dict__["st_size"] = 4 * 1024
        c.setattr(Path, 'stat', lambda _: obj)
        c.setattr(Path, 'read_text', lambda _: simple_changelists_xml)
        result = load_storage(None)
    assert len(result.get_changelists()) == 1


def test_load_storage_none_workspace_simple_cl_raises_exit(simple_workspace_xml):
    """ This case raises exit because it checks Changelists default path first.
    Due to MP, it sees that every file exists, but the provider gives Workspace XML format.
    Since it expects Changelists XML, it will exit after parsing xml and searching for changelists tag.
    """
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'is_file', lambda _: True)
        obj = Mock()
        obj.__dict__["st_size"] = 4 * 1024
        c.setattr(Path, 'stat', lambda _: obj)
        c.setattr(Path, 'read_text', lambda _: simple_workspace_xml)
        try:
            result = load_storage(None)
            assert len(result.get_changelists()) == 1
            raises_exit = False
        except SystemExit:
            raises_exit = True
        assert raises_exit


def test_load_storage_none_with_file_path_raises_exit(temp_file):
    try:
        result = load_storage(None, temp_file)
        assert len(result.get_changelists()) == 1
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit



""" Testing the ChangelistsTree Class.
"""
from changelist_data.xml import changelists


def test_extract_list_elements_simple_returns_list(simple_cl_tree, simple_cl):
    result = simple_cl_tree.get_changelists()
    assert len(result) == 1
    assert simple_cl == result[0]


def test_extract_list_elements_multi_returns_list(multi_cl_tree, multi_cl_list):
    result = multi_cl_tree.get_changelists()
    assert len(result) == 2
    assert multi_cl_list == result


def test_extract_list_elements_no_cl_returns_empty_list(no_cl_tree):
    assert len(no_cl_tree.get_changelists()) == 0


def test_update_changelists_simple_with_empty(simple_cl_tree):
    simple_cl_tree.update_changelists([])
    assert len(simple_cl_tree.get_changelists()) == 0


def test_update_changelists_simple_with_multi(simple_cl_tree, multi_cl_tree):
    simple_cl_tree.update_changelists(
        multi_cl_tree.get_changelists()
    )
    result = simple_cl_tree.get_changelists()
    assert len(result) == 2
    assert result == multi_cl_tree.get_changelists()


def test_update_changelists_multi_with_simple(multi_cl_tree, simple_cl_tree):
    multi_cl_tree.update_changelists(
        simple_cl_tree.get_changelists()
    )
    result = multi_cl_tree.get_changelists()
    assert len(result) == 1
    assert result == simple_cl_tree.get_changelists()


def test_changelists_tree_no_changelists_tag_returns_empty_list(empty_changelists_xml):
    tree = changelists.load_xml(empty_changelists_xml)
    result = tree.get_changelists()
    assert len(result) == 0


def test_get_root_no_cl_manager_returns_root(no_cl_tree):
    assert no_cl_tree.get_root().getroot().tag == 'changelists'


def test_get_root_simple_returns_root(simple_cl_tree):
    assert simple_cl_tree.get_root().getroot().tag == 'changelists'


def test_get_root_multi_returns_root(multi_cl_tree):
    assert multi_cl_tree.get_root().getroot().tag == 'changelists'



"""
"""
import tempfile
from pathlib import Path

import pytest


@pytest.fixture
def temp_dir():
    tdir = tempfile.TemporaryDirectory()
    yield tdir
    tdir.cleanup()


@pytest.fixture
def temp_file(temp_dir):
    file = temp_dir.name + "/test_xml.xml"
    yield Path(file)



""" Testing Workspace XML Package Methods.
"""
from changelist_data.xml.workspace import load_xml


def test_load_xml_empty_str_raises_exit():
    try:
        load_xml("")
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit


def test_load_xml_no_cl_returns_new_tree(no_clm_workspace_xml):
    try:
        result = load_xml(no_clm_workspace_xml)
        result.get_changelists()
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit


def test_load_xml_simple_cl(simple_workspace_xml, simple_ws_tree):
    assert simple_ws_tree.get_changelists() == load_xml(simple_workspace_xml).get_changelists()


def test_load_xml_multi_cl(multi_workspace_xml, multi_ws_tree):
    assert multi_ws_tree.get_changelists() == load_xml(multi_workspace_xml).get_changelists()



""" Testing Changelists XML read_xml Method.
"""
from changelist_data.xml.changelists import read_xml


def test_read_xml_empty_str_raises_exit():
    try:
        read_xml("")
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit


def test_read_xml_no_cl(empty_changelists_xml):
    result = read_xml(empty_changelists_xml)
    assert len(result) == 0


def test_read_xml_simple_cl(simple_changelists_xml, simple_cl_list):
    result = read_xml(simple_changelists_xml)
    assert len(result) == 1
    assert simple_cl_list == result


def test_read_xml_multi_cl(multi_changelists_xml, multi_cl_list):
    result = read_xml(multi_changelists_xml)
    assert len(result) == 2
    assert multi_cl_list == result



"""Testing Changelist Reader Methods.
"""
import pytest

from changelist_data.xml.workspace.workspace_reader import find_changelist_manager, parse_xml, extract_list_elements


@pytest.fixture()
def simple_clm(simple_workspace_xml):
    return find_changelist_manager(parse_xml(simple_workspace_xml))


@pytest.fixture()
def multi_clm(multi_workspace_xml):
    return find_changelist_manager(parse_xml(multi_workspace_xml))


def test_find_changelist_manager_empty_xml_raises_exit():
    try:
        find_changelist_manager(parse_xml(""))
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit


def test_find_changelist_manager_no_changelist_returns_none(no_clm_workspace_xml):
    assert find_changelist_manager(parse_xml(no_clm_workspace_xml)) is None


def test_find_changelist_manager_simple_changelist_returns_element(simple_workspace_xml):
    element = find_changelist_manager(parse_xml(simple_workspace_xml))
    change_lists = list(element.iter())
    assert len(change_lists) == 3


def test_find_changelist_manager_multi_changelist_returns_element(multi_workspace_xml):
    element = find_changelist_manager(parse_xml(multi_workspace_xml))
    change_lists = list(element.iter())
    assert len(change_lists) == 6


def test_extract_list_elements_simple_clm_(simple_clm, simple_cl):
    elem = extract_list_elements(simple_clm)
    assert len(elem) == 1
    assert simple_cl == elem[0]


def test_extract_list_elements_multi_clm_(multi_clm, multi_cl_list):
    assert multi_cl_list == extract_list_elements(multi_clm)



"""Setup Package Configuration
"""
from setuptools import setup, find_packages


setup(
    name="changelist-data",
    version="0.1.4",
	description='Data Management base package for Changelists CLI Tools',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
	author='DK96-OS',
	url='https://github.com/DK96-OS/changelist-data/',
	project_urls={
        "Issues": "https://github.com/DK96-OS/changelist-data/issues",
        "Source Code": "https://github.com/DK96-OS/changelist-data/"
	},
	license='GPLv3',
    packages=find_packages(exclude=['test']),
    entry_points={
        'console_scripts': [],
    },
    python_requires='>=3.10',
    keywords=['changelist', 'vcs'],
    classifiers=[
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
)



""" Changelist Data Storage XML Workspace
"""
from changelist_data.changelist import Changelist
from changelist_data.xml.workspace import workspace_reader
from changelist_data.xml.workspace.workspace_tree import WorkspaceTree


def read_xml(
    workspace_xml: str
) -> list[Changelist]:
    """
    Parse the Workspace XML file and obtain all ChangeList Data in a list.

    Parameters:
    - workspace_xml (str): The contents of the Workspace file, in xml format.
    
    Returns:
    list[Changelist] - The list of Changelists in the workspace file.
    """
    if (cl_manager := workspace_reader.find_changelist_manager(workspace_reader.parse_xml(workspace_xml))) is None:
        exit("ChangeList Manager was not found in the workspace file.")
    return workspace_reader.extract_list_elements(cl_manager)


def load_xml(
    workspace_xml: str
) -> WorkspaceTree:
    """
    Parse the Workspace XML file into an XML Tree, and Wrap it.

    Parameters:
    - workspace_xml (str): The contents of the Workspace file, in xml format.

    Returns:
    WorkspaceTree - An XML Tree changelists interface.
    """
    return WorkspaceTree(
        workspace_reader.parse_xml(workspace_xml)
    )



""" Testing BaseXML Tree
"""
import tempfile
from pathlib import Path

import pytest

from changelist_data.xml import changelists, workspace


@pytest.fixture
def temp_dir():
    tdir = tempfile.TemporaryDirectory()
    yield tdir
    tdir.cleanup()


@pytest.fixture
def temp_file(temp_dir):
    file = temp_dir.name + "/test_xml.xml"
    yield Path(file)


def test_write_file_changelists_new_tree_empty(temp_file):
    tree = changelists.new_tree()
    tree.write_tree(temp_file)
    # Read from File
    result = changelists.read_xml(temp_file.read_text())
    assert len(result) == 0


def test_write_file_workspace_simple_cl(temp_file, simple_workspace_xml):
    tree = workspace.load_xml(simple_workspace_xml)
    tree.write_tree(temp_file)
    # Read from File
    result = workspace.read_xml(temp_file.read_text())
    assert len(result) == 1


def test_write_file_changelists_simple_cl(temp_file, simple_changelists_xml):
    tree = changelists.load_xml(simple_changelists_xml)
    tree.write_tree(temp_file)
    # Read from File
    result = changelists.read_xml(temp_file.read_text())
    assert len(result) == 1



""" Testing Storage of Workspace XML Files.
"""
from changelist_data.storage import changelists_storage, storage_type, StorageType
from changelist_data.xml.changelists import new_tree


def test_read_file_empty_returns_new_tree(temp_file):
    temp_file.write_text("")
    result = changelists_storage.read_file(temp_file)
    assert len(result) == 0


def test_read_file_invalid_xml_raises_exit(temp_file):
    temp_file.write_text("invalid xml")
    try:
        changelists_storage.read_file(temp_file)
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit


def test_read_file_none_returns_empty_list():
    result = changelists_storage.read_file(None)
    assert len(result) == 0


def test_read_file_no_cl(temp_file, empty_changelists_xml):
    temp_file.write_text(empty_changelists_xml)
    result = changelists_storage.read_file(temp_file)
    assert len(result) == 0


def test_read_file_simple_cl(temp_file, simple_changelists_xml):
    temp_file.write_text(simple_changelists_xml)
    result = changelists_storage.read_file(temp_file)
    assert len(result) == 1


def test_read_file_multi_cl(temp_file, multi_changelists_xml):
    temp_file.write_text(multi_changelists_xml)
    result = changelists_storage.read_file(temp_file)
    assert len(result) == 2


def test_load_file_invalid_xml_raises_exit(temp_file):
    temp_file.write_text("invalid xml")
    try:
        changelists_storage.load_file(temp_file)
        raises_exit = False
    except SystemExit:
        raises_exit = True
    assert raises_exit


def test_load_file_none_returns_new_tree(temp_file, empty_changelists_xml):
    temp_file.write_text(empty_changelists_xml)
    result = changelists_storage.load_file(None)
    assert result.get_changelists() == new_tree().get_changelists()


def test_load_file_no_cl(temp_file, empty_changelists_xml):
    temp_file.write_text(empty_changelists_xml)
    result = changelists_storage.load_file(temp_file)
    assert len(result.get_changelists()) == 0


def test_load_file_simple_cl(temp_file, simple_changelists_xml):
    temp_file.write_text(simple_changelists_xml)
    result = changelists_storage.load_file(temp_file)
    assert len(result.get_changelists()) == 1


def test_load_file_multi_cl(temp_file, multi_changelists_xml):
    temp_file.write_text(multi_changelists_xml)
    result = changelists_storage.load_file(temp_file)
    assert len(result.get_changelists()) == 2


def test_write_file_none_tree(temp_file):
    temp_file.write_text("")
    changelists_storage.write_file(None, temp_file)
    # Read from Temp
    result = changelists_storage.read_file(temp_file)
    assert len(result) == 0


def test_write_file_no_cl(temp_file, empty_changelists_xml):
    temp_file.write_text(empty_changelists_xml)
    tree = changelists_storage.load_file(temp_file)
    changelists_storage.write_file(tree, temp_file)
    # Read from Temp
    result = changelists_storage.read_file(temp_file)
    assert len(result) == 0


def test_write_file_simple_cl(temp_file, simple_changelists_xml):
    temp_file.write_text(simple_changelists_xml)
    tree = changelists_storage.load_file(temp_file)
    changelists_storage.write_file(tree, temp_file)
    # Read from Temp
    result = changelists_storage.read_file(temp_file)
    assert len(result) == 1


def test_write_file_multi_cl(temp_file, multi_changelists_xml):
    temp_file.write_text(multi_changelists_xml)
    tree = changelists_storage.load_file(temp_file)
    changelists_storage.write_file(tree, temp_file)
    # Read from Temp
    result = changelists_storage.read_file(temp_file)
    assert len(result) == 2


def test_write_file_empty_default_changelist(temp_file):
    temp_file.write_text('')
    #
    tree = changelists_storage.load_file(
        storage_type.get_default_path(StorageType.CHANGELISTS)
    )
    changelists_storage.write_file(tree, temp_file)
    # Read from Temp
    result = changelists_storage.read_file(temp_file)
    assert len(result) == 0


def test_write_file_new_tree_empty():
    tree = new_tree()
    tree.write_tree(
        storage_type.get_default_path(StorageType.CHANGELISTS)
    )
    # Read from File
    result = changelists_storage.read_file(
        storage_type.get_default_path(StorageType.CHANGELISTS)
    )
    assert len(result) == 0



""" An Abstract Class defining the interface for translation between XML Trees and Changelists.
"""
from dataclasses import dataclass
from pathlib import Path

from changelist_data.changelist import Changelist
from changelist_data.storage.storage_type import StorageType
from changelist_data.xml.base_xml_tree import BaseXMLTree


@dataclass(frozen=True)
class ChangelistDataStorage:
    """ Controller Interface for Data Storage.

    Fields:
    - base_xml_tree (BaseXMLTree): A Wrapper around XML ElementTree, compatible with both xml format Storage Types.
    - storage_type (StorageType): An enum selecting the specific xml format used.
    - update_path (Path): The file Path where Data is read from and written to.
    """
    base_xml_tree: BaseXMLTree
    storage_type: StorageType
    update_path: Path

    def get_changelists(self) -> list[Changelist]:
        return self.base_xml_tree.get_changelists()

    def update_changelists(
        self, changelists: list[Changelist]
    ):
        """ Overwrite the collection of Changelist data in Memory.

        Parameters:
        - changelists (list[Changelist]): The new list of Changelists.
        """
        self.base_xml_tree.update_changelists(changelists)

    def write_to_storage(self) -> bool:
        """ Create or overwrite storage file.
            Ensures parent directories exist.

        Returns:
        bool - True if data was written.
        """
        return self.base_xml_tree.write_tree(self.update_path)



"""XML Reader Methods.
"""
from typing import Iterator
from xml.etree.ElementTree import Element


def filter_by_tag(element: Element, tag: str) -> Iterator[Element]:
    """
    Obtain all Elements that match a given Tag.
    
    Parameters:
    - element (Element): The element to search within.
    - tag (str): The name of the Tag to filter for.

    Returns:
    list[Element]
    """
    return filter(lambda x: x.tag == tag, element.iter())


def get_attr(element: Element, attr: str) -> str | None:
    """
    Obtain the Attribute from an element safely, returning None.

    Parameters:
    - element (Element): The element to obtain the attribute from.
    - attr (str): The name of the attribute.

    Returns:
    str | None - The value of the attribute, or None.
    """
    return element.attrib[attr] if attr in element.attrib else None


def get_attr_or(element: Element, attr: str, default: str) -> str:
    """
    Obtain the Attribute from an element safely, returning None if not found.

    Parameters:
    - element (Element): The element to obtain the attribute from.
    - attr (str): The name of the attribute.

    Returns:
    str - The value of the attribute in the element, or the default.
    """
    return element.attrib[attr] if attr in element.attrib else default


def read_bool_from(element: Element, attr: str) -> bool:
    """
    Obtain the Attribute from an element safely, returning None if not found.

    Parameters:
    - element (Element): The element to obtain the attribute from.
    - attr (str): The name of the attribute.

    Returns:
    bool - The attribute converted to a boolean.
    """
    value = get_attr_or(element, attr, 'false')
    return value.lower() == 'true'


def convert_bool(attr: str | None) -> bool | None:
    """
    Convert a String attribute to boolean.
    """
    if attr is None:
        return None
    return attr == 'true'


"""The Data Class for a ChangeList.
"""
from dataclasses import dataclass

from changelist_data.file_change import FileChange


@dataclass(frozen=True)
class Changelist:
    """
    The Data class representing a ChangeList.
    
    Properties:
    - id (str): The unique id of the changelist.
    - name (str): The name of the changelist.
    - changes (list[ChangeData]): The list of file changes in the changelist.
    - comment (str): The comment associated with the changelist.
    - is_default (bool): Whether this is the active changelist.
    """
    id: str
    name: str
    changes: list[FileChange]
    comment: str = ""
    is_default: bool = False



"""Testing XML Reader Methods.
"""
from xml.etree.ElementTree import fromstring

from changelist_data.xml.xml_reader import filter_by_tag


def test_filter_by_tag_no_changelist_returns_empty(empty_changelists_xml):
    root_elem = fromstring(empty_changelists_xml)
    # There are no List Elements
    assert len(list(filter_by_tag(root_elem, 'list'))) == 0
    # There are no Change Elements
    assert len(list(filter_by_tag(root_elem, 'change'))) == 0


def test_filter_by_tag_simple_returns_1_list_1_change(simple_changelists_xml):
    root_elem = fromstring(simple_changelists_xml)
    # There is 1 List Element
    assert len(list(filter_by_tag(root_elem, 'list'))) == 1
    # There is 1 Change Element
    assert len(list(filter_by_tag(root_elem, 'change'))) == 1


def test_filter_by_tag_multi_returns_2_list_3_changes(multi_changelists_xml):
    root_elem = fromstring(multi_changelists_xml)
    # There are 2 List Elements
    assert len(list(filter_by_tag(root_elem, 'list'))) == 2
    # There are 3 Change Elements
    assert len(list(filter_by_tag(root_elem, 'change'))) == 3



""" XML Workspace Trees
"""
import pytest

from changelist_data.xml import workspace


@pytest.fixture()
def simple_ws_tree(simple_workspace_xml):
    return workspace.load_xml(simple_workspace_xml)


@pytest.fixture()
def multi_ws_tree(multi_workspace_xml):
    return workspace.load_xml(multi_workspace_xml)


@pytest.fixture()
def no_clm_ws_tree(no_clm_workspace_xml):
    return workspace.load_xml(no_clm_workspace_xml)



""" Methods for Writing XML.
"""
from xml.etree.ElementTree import Element, indent

from changelist_data.file_change import FileChange
from changelist_data.changelist import Changelist
from changelist_data.xml.workspace import workspace_reader


def write_list_element(
    changelist: Changelist,
    indent_level: int = 2,
) -> Element:
    """
    Convert a Changelist to XML Element format.

    Parameters:
    - changelist (Changelist): The Changelist data to format.

    Returns:
    str - A String containing the xml formatted contents of the Changelist. 
    """
    clist = Element('list')
    if changelist.is_default:
        clist.set('default', 'true')
    clist.set('id', changelist.id)
    clist.set('name', changelist.name)
    clist.set('comment', changelist.comment)
    for change in changelist.changes:
        clist.append(
            _write_change_data(
                data=change,
                indent_level=indent_level + 1,
            )
        )
    indent(clist, level=indent_level)
    return clist


def _write_change_data(
    data: FileChange,
    indent_level: int,
) -> Element:
    """
    Write the FileChange Data to XML format.

    Parameters:
    - data (FileChange): The FileChange Data for a specific File.

    Returns:
    str - A String containing the XML tag for this FileChange Data.
    """
    change = Element('change')
    if data.before_path is not None:
        change.set('beforePath', workspace_reader._PROJECT_DIR_VAR + data.before_path)
    if data.before_dir is not None:
        change.set('beforeDir', str(data.before_dir).lower())
    if data.after_path is not None:
        change.set('afterPath',workspace_reader._PROJECT_DIR_VAR + data.after_path)
    if data.after_dir is not None:
        change.set('afterDir', str(data.after_dir).lower())
    #
    indent(change, level=indent_level)
    return change
